"""
A Python script to convert excel files into JSON.
"""

import json
import re
import sys
import codecs
import os
import constants
from errors import PyXFormError
from xls2json_backends import xls_to_dict, csv_to_dict



# the following are the three sheet names that this program expects
SURVEY_SHEET = u"survey"
CHOICES = u"choices"
SETTINGS = u"settings"
CHOICES_SHEET_NAMES = [u"choices", u"choices and columns"]
COLUMNS = u"columns"

COLUMNS_SHEET = u"columns" #Not used
TYPES_SHEET = u"question types"

LIST_NAME = u"list name"
# Special reserved values for type column that allow the user to set
# the form's title or id.
SET_TITLE = u"set form title"
SET_ID = u"set form id"
SET_DEFAULT_LANG = u"set default language"

group_name_conversions = {
    "looped group": u"repeat"
}
sheet_name_conversions = {
    u"choices" : u"lists",
    u"choices and columns" : u"lists",
    u"columns" : u"lists"
}

def print_pyobj_to_json(pyobj, path):
    fp = codecs.open(path, mode="w", encoding="utf-8")
    json.dump(pyobj, fp=fp, ensure_ascii=False, indent=4)
    fp.close()

#Actually I shouldn't separate this out because I will also want to apply various transformations to the sheet before and during validation.
def validate_spreadsheet(spreadsheet_dict):
    """
    spreadsheet_dict -- a dealiased spreadsheet dictionary
    I'm choosing to validate at the spreadsheet stage as opposed to the json or xform stage
    because it will enable us to give the clearest feedback to the user. (i.e. say "row 19 has a duplicate name")
    """
    if SURVEY_SHEET not in spreadsheet_dict:
        raise PyXFormError("You must have a sheet named: " + SURVEY_SHEET)
    #Check for duplicate names
    name_set = set()
    for sheet_name, dicts in spreadsheet_dict.items():
        for dicty in dicts:
            if constants.NAME not in dicty:
                #TODO: Warn
                continue
            if dicty[constants.NAME] in name_set:
                raise PyXFormError("Duplicate name: " + dicty[constants.NAME])
                    
def group_dictionaries(spreadsheet_dict):
    """
    For each row in the worksheet, group all keys that contain a
    colon. So {"text:english": "hello", "text:french" :
    "bonjour"} becomes {"text": {"english": "hello", "french" :
    "bonjour"}.
    """
    DICT_CHAR = u":"
    for sheet_name, dicts in spreadsheet_dict.items():
        for dicty in dicts:
            groups = {}
            for k, v in dicty.items():
                l = k.split(DICT_CHAR)
                if len(l) >= 2:
                    if l[0] not in groups:
                        groups[l[0]] = {}
                    groups[l[0]][DICT_CHAR.join(l[1:])] = v
                    del dicty[k]
            for k, v in groups.items():
                assert k not in dicty
                dicty[k] = v

class SpreadsheetReader(object):
    
    def __init__(self, path_or_file):
        if isinstance(path_or_file, basestring):
            self._file_object = None
            path = path_or_file
        else:
            self._file_object = path_or_file
            path = self._file_object.name

        (filepath, filename) = os.path.split(path)
        (shortname, extension) = os.path.splitext(filename)
        self.filetype = None
        if extension == ".xlsx":
            raise PyXFormError("XLSX files are not supported at this time. Please save the spreadsheet as an XLS file (97).")
        elif extension == ".xls":
            self.filetype = "xls"
        elif extension == ".csv":
            self.filetype = "csv"
        self._path = path
        self._name = unicode(shortname)
        self._print_name = unicode(shortname)
        self._title = unicode(shortname)
        self._id = unicode(shortname)
        self._def_lang = unicode("English")
        self._parse_input()

    def _parse_input(self):
        if self.filetype == None:
            raise PyXFormError("File was not recognized")
        elif self.filetype == "xls":
            self._dict = xls_to_dict(self._file_object if self._file_object is not None else self._path)
        elif self.filetype == "csv":
            self._dict = csv_to_dict(self._file_object if self._file_object is not None else self._path)
        self._sheet_names = self._dict.keys()
        self._set_choices_and_columns_sheet_name()
        self._strip_unicode_values()
        self._fix_int_values()
        group_dictionaries(self._dict)

    def _set_choices_and_columns_sheet_name(self):
        """
        If the xls file has a sheet with a name in CHOICES_SHEET_NAMES
        _lists_sheet_name is set to it.
        """
        sheet_names = self._dict.keys()

        self._lists_sheet_name = None
        for sheet_name in sheet_names:
            if sheet_name in CHOICES_SHEET_NAMES:
                self._lists_sheet_name = sheet_name

    def _strip_unicode_values(self):
        for sheet_name, dicts in self._dict.items():
            for d in dicts:
                for k, v in d.items():
                    if type(v) == unicode:
                        d[k] = v.strip()

    def _fix_int_values(self):
        """
        Excel only has floats, but we really want integer values to be
        ints.
        """
        for sheet_name, dicts in self._dict.items():
            for d in dicts:
                for k, v in d.items():
                    if type(v) == float and v == int(v):
                        d[k] = int(v)

    def to_json_dict(self):
        return self._dict

    #TODO: Make sure the unicode chars don't show up
    def print_json_to_file(self, filename=""):
        if not filename:
            filename = self._path[:-4] + ".json"
        print_pyobj_to_json(self.to_json_dict(), filename)

#Aliases
yes_no_conversions = {
    "yes": "true()",
    "Yes": "true()",
    "YES": "true()",
    "true": "true()",
    "True": "true()",
    "TRUE": "true()",
    "no": "false()",
    "No": "false()",
    "NO": "false()",
    "false": "false()",
    "False": "false()",
    "FALSE": "false()"
}
control_conversions = {
    u"group" : u"group",
    u"lgroup" : u"group",
    u"repeat": u"repeat",
    u"loop": u"repeat",
    u"looped group": u"repeat"
}
def dealias_headers(dict_array):
    """
    Copies dict_array so this isn't super efficient.
    """
    #TODO: Check on bind prefix.
    column_header_conversions = {
        u"constraint_message" : u"constraint message",
        u"read_only" : u"read only", 
        u"select_one" : u"select one",
        u"select_multiple" : u"select multiple",
        u"list_name" : u"list name"
    }
    out_dict_array = list()
    for row in dict_array:
        out_row = dict()
        for key in row.keys():
            if key in column_header_conversions.keys():
                out_row[column_header_conversions[key]] = row[key]
            else:
                out_row[key] = row[key]
        out_dict_array.append(out_row)
    return out_dict_array

class ParseQuestionException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)
    
class SurveyReader(SpreadsheetReader):

    def __init__(self, path):
        super(SurveyReader, self).__init__(path)
        self._setup_survey()

    def _setup_survey(self):
        """
        Does some parsing on the survey sheet.
        I think this should probably go somewhere else so all the parsing happens in the same place.
        """
        self._process_questions()
        self._construct_choice_lists()
        self._insert_lists()
        self._save_settings()
        self._organize_sections()

    def _save_settings(self):
        """
        sets _settings to the settings worksheet.
        """
        # the excel reader gives a list of dicts, one dict for each
        # row after the headers, we're only going to use the first
        # row.
        self._settings = self._dict.get(u"settings", [{}])[0]

    def _process_questions(self):
        """
        A question is dictionary representing a row in the survey worksheet where the keys are the column headers.
        This function does some light parsing on them, for example it will:
        remove disabled questions, set the survey title and id, break apart select and group statements.
        """
        new_question_list = list()
        self._dict[SURVEY_SHEET] = dealias_headers(self._dict[SURVEY_SHEET])
        for question in self._dict[SURVEY_SHEET]:
            if u"type" not in question:
                continue
            #Disabled should probably be first so the attributes below can be disabled.
            if u"disabled" in question:
                disabled = question[u"disabled"]
                if disabled in yes_no_conversions:
                    disabled = yes_no_conversions[disabled]
                if disabled == 'true()':
                    continue
                
            #TODO: These should be on the settings sheet... I'm not sure if we need to support them being on the survey sheet as well
            #    Except default lang, I don't know what to do with that.
            if question[constants.TYPE] == SET_TITLE:
                if not question[constants.NAME].strip().find(" ") == -1:
                    raise PyXFormError("Form title must not include any spaces", question[constants.NAME])
                self._title = question[constants.NAME]
                continue
    
            if question[constants.TYPE] == SET_ID:
                #TODO: Can any name cell ever contain spaces? Move this up if not
                if not question[constants.NAME].strip().find(" ") == -1:
                    raise PyXFormError("Form id must not include any spaces", question[constants.NAME])
                self._id = question[constants.NAME]
                continue
            
            if question[constants.TYPE] == SET_DEFAULT_LANG:
                self._def_lang = question[constants.NAME]#We need to hold onto this because it is used when generating itext elements
                continue
            
            new_question_list.append(self._process_question_type(question))
            
        self._dict[SURVEY_SHEET] = new_question_list
        
        #Make sure form name and ID are properly set:
        if self._id.find(" ") != -1:
            raise PyXFormError("Form id must not include any spaces", self._id)

        if self._name.find(" ") != -1:
            self._name = self._id

    def _process_question_type(self, question):
        question_type = question[constants.TYPE]
        question_type.strip()
        question_type = re.sub(r"\s+", " ", question_type) #Remove double spaces?

        try:
            return self._prepare_multiple_choice_question(question, question_type)
        except ParseQuestionException:
            try:
                return self._prepare_begin_loop(question, question_type)
            except ParseQuestionException as e:
                #print e.value #just for debug, maybe this should print to a logfile
                #raise PyXFormError("Unsupported syntax: '%s'" % question_type)
                return question

    def _prepare_multiple_choice_question(self, question, question_type):
        """
        Parse a multple choice question
        Throws ParseQuestionException
        Returns the passed in reference to the question object
        """
        selectCommands = {#Old commands
                          "select all that apply from" : u"select all that apply",
                          "select one from" : u"select one", 
                          #New commands
                          "select_one" : u"select one",
                          "select_multiple" : u"select all that apply" }
        
        select_regexp = r"^(?P<select_command>(" + '|'.join(selectCommands.keys()) + r")) (?P<list_name>\S+)( (?P<specify_other>or specify other))?$"
        select_parse = re.search(select_regexp, question_type)
        if select_parse:
            parse_dict = select_parse.groupdict()
            if parse_dict["select_command"]:
                select_type = selectCommands[parse_dict["select_command"]]
                list_name = parse_dict["list_name"] #TODO: should check that this is valid at some point
                
                #TODO: specify_other is not in the new spec
                specify_other = ("specify_other" in parse_dict and parse_dict["specify_other"]) or (" or specify other" in parse_dict and parse_dict[" or specify other"]) #old version

                question[constants.TYPE] = select_type
                if specify_other:
                    question[constants.TYPE] += " or specify other"
                question[CHOICES] = list_name
                
                return question
                
        raise ParseQuestionException("")

    def _prepare_begin_loop(self, q, question_type):
        m = re.search(r"^(?P<type>begin loop) over (?P<list_name>\S+)$", question_type)
        if not m:
            raise ParseQuestionException("Regex search returned None")
            #raise PyXFormError("unsupported loop syntax:" + question_type)
        assert COLUMNS not in q
        d = m.groupdict()
        q[COLUMNS] = d["list_name"]
        q[constants.TYPE] = d["type"]
        return q

    def _construct_choice_lists(self):
        """
        Each choice has a list name associated with it. Go through the
        list of choices, grouping all the choices by their list name.
        """
        if self._lists_sheet_name is None:
            return
        choice_list = self._dict[self._lists_sheet_name]
        choices = {}
        for choice in choice_list:
            try:
                #TODO: decide whether there should be an underscore in list_name so we can get rid of this.
                list_name_string_wo_underscore = re.sub(" ", "_", LIST_NAME) if LIST_NAME not in choice else LIST_NAME
                list_name = choice.pop(list_name_string_wo_underscore)
                if list_name in choices:
                    choices[list_name].append(choice)
                else:
                    choices[list_name] = [choice]
            except KeyError:
                raise PyXFormError("For some reason this choice isn't associated with a list.", choice)
        self._dict[self._lists_sheet_name] = choices

    def _insert_lists(self):
        """
        For each multiple choice question and loop in the survey find
        the corresponding list and add it to that question.
        """
        lists_by_name = self._dict.get(self._lists_sheet_name, {})
        for q in self._dict[SURVEY_SHEET]:
            self._insert_list(q, CHOICES, lists_by_name)
            self._insert_list(q, COLUMNS, lists_by_name)

    def _insert_list(self, q, key, lists_by_name):
        if key in q:
            list_name = q[key]
            if list_name not in lists_by_name:
                raise PyXFormError("There is no list of %s by this name" % key, list_name)
            q[key] = lists_by_name[list_name]

    def _organize_sections(self):
        """
        This function arranges all the sections into a tree structure
        """
        # this needs to happen after columns have been inserted
        self._dict = self._dict[SURVEY_SHEET]
        result = {u"type": u"survey", u"name": self._name, u"children": []}
        result.update(self._settings)
        stack = [result]
        for cmd in self._dict:
            cmd_type = cmd[u"type"]
            match_begin = re.match(r"begin (?P<type>group|repeat|loop)", cmd_type)
            match_end = re.match(r"end (?P<type>group|repeat|loop)", cmd_type)
            # TODO: combine the begin and end patterns below with those above.
            # match_begin = re.match(r"begin (?P<type>lgroup|group|looped group|repeat)", cmd_type)
            # match_end = re.match(r"end (?P<type>lgroup|group|looped group|repeat)", cmd_type)
            if match_begin:
                # start a new section
                cmd[u"type"] = match_begin.group(1)

                if cmd[u"type"] in group_name_conversions:
                    cmd[u"type"] = group_name_conversions[cmd[u"type"]]

                cmd[u"children"] = []
                stack[-1][u"children"].append(cmd)
                stack.append(cmd)
            elif match_end:
                match_end = match_end.group(1)
                if match_end in group_name_conversions:
                    match_end = group_name_conversions[match_end]

                begin_cmd = stack.pop()
                if begin_cmd[u"type"] != match_end:
                    raise PyXFormError("This end group does not match the previous begin", cmd)
            else:
                stack[-1][u"children"].append(cmd)
        self._dict = result


class QuestionTypesReader(SpreadsheetReader):
    """
    Class for reading spreadsheet file that specifies the available question types.
    @see question_type_dictionary
    """
    def __init__(self, path):
        super(QuestionTypesReader, self).__init__(path)
        self._setup_question_types_dictionary()

    def _setup_question_types_dictionary(self):
        self._dict = self._dict[TYPES_SHEET]
        self._organize_by_type_name()

    def _organize_by_type_name(self):
        result = {}
        for question_type in self._dict:
            result[question_type.pop(u"name")] = question_type
        self._dict = result


#Not used internally
class VariableNameReader(SpreadsheetReader):
    def __init__(self, path):
        SpreadsheetReader.__init__(self, path)
        self._organize_renames()

    def _organize_renames(self):
        new_dict = {}
        variable_names_so_far = []
        assert u"Dictionary" in self._dict
        for d in self._dict[u"Dictionary"]:
            if u"Variable Name" in d:
                assert d[u"Variable Name"] not in variable_names_so_far, \
                    d[u"Variable Name"]
                variable_names_so_far.append(d[u"Variable Name"])
                new_dict[d[u"XPath"]] = d[u"Variable Name"]
            else:
                variable_names_so_far.append(d[u"XPath"])
        self._dict = new_dict


if __name__ == "__main__":
    # Open the excel file that is the second argument to this python
    # call, convert that file to json and save that json to a file
    path = sys.argv[1]
    converter = SurveyReader(path)
    # converter.print_json_to_file()
    print json.dumps(converter.to_json_dict(), ensure_ascii=False, indent=4)



import tempfile
from section import Section
from question import Question
from utils import node, XFORM_TAG_REGEXP
from collections import defaultdict
import codecs
from datetime import datetime
import re
import os
from odk_validate import check_xform
from survey_element import SurveyElement
from errors import PyXFormError
from pyxform import constants


nsmap = {
    u"xmlns": u"http://www.w3.org/2002/xforms",
    u"xmlns:h": u"http://www.w3.org/1999/xhtml",
    u"xmlns:ev": u"http://www.w3.org/2001/xml-events",
    u"xmlns:xsd": u"http://www.w3.org/2001/XMLSchema",
    u"xmlns:jr": u"http://openrosa.org/javarosa",
    u"xmlns:orx": u"http://openrosa.org/xforms/"
    }

class Survey(Section):

    FIELDS = Section.FIELDS.copy()
    FIELDS.update(
        {
            u"_xpath": dict,
            u"_created": datetime.now, #This can't be dumped to json
            u"title": unicode,
            u"id_string": unicode,
            u"sms_keyword": unicode,
            u"sms_separator": unicode,
            u"sms_allow_media": bool,
            u"sms_date_format": unicode,
            u"sms_datetime_format": unicode,
            u"sms_response": unicode,
            u"file_name": unicode,
            u"default_language": unicode,
            u"_translations": dict,
            u"submission_url": unicode,
            u"public_key": unicode,
            u"instance_xmlns": unicode,
            u"version": unicode,
            u"choices": dict,
            u"style": unicode
        }
    )

    def validate(self):
        super(Survey, self).validate()
        self._validate_uniqueness_of_section_names()

    def _validate_uniqueness_of_section_names(self):
        section_names = []
        for e in self.iter_descendants():
            if isinstance(e, Section):
                if e.name in section_names:
                    raise PyXFormError("There are two sections with the name %s." % e.name)
                section_names.append(e.name)

    def xml(self):
        """
        calls necessary preparation methods, then returns the xml.
        """
        self.validate()
        self._setup_xpath_dictionary()
        body_kwargs = {}
        if hasattr(self, constants.STYLE) and getattr(
                self, constants.STYLE):
            body_kwargs['class'] = getattr(
                self, constants.STYLE)
        return node(u"h:html",
                    node(u"h:head",
                         node(u"h:title", self.title),
                         self.xml_model()
                        ),
                    node(u"h:body", *self.xml_control(), **body_kwargs),
                    **nsmap
                    )

    def _generate_static_instances(self):
        """
        Generates <instance> elements for static data (e.g. choices for select type questions)
        """
        for list_name, choice_list in self.choices.items():
            instance_element_list = []
            for idx, choice in zip(range(len(choice_list)), choice_list):
                choice_element_list = []
                #Add a unique id to the choice element incase there is itext it refrences
                itextId = '-'.join(['static_instance', list_name, str(idx)])
                choice_element_list.append(node("itextId", itextId))

                for choicePropertyName, choicePropertyValue in choice.items():
                    if isinstance(choicePropertyValue, basestring) and choicePropertyName != 'label':
                        choice_element_list.append(node(choicePropertyName, unicode(choicePropertyValue)))
                instance_element_list.append(node("item", *choice_element_list))
            yield node("instance", node("root", *instance_element_list), id=list_name)

    def xml_model(self):
        """
        Generate the xform <model> element
        """
        self._setup_translations()
        self._setup_media()
        self._add_empty_translations()

        model_children = []
        if self._translations:
            model_children.append(self.itext())
        model_children += [node("instance", self.xml_instance())]
        model_children += list(self._generate_static_instances())
        model_children += self.xml_bindings()

        if self.submission_url or self.public_key:
            submission_attrs = dict()
            if self.submission_url:
                submission_attrs["action"] = self.submission_url
            if self.public_key:
                submission_attrs["base64RsaPublicKey"] = self.public_key
            submission_node = node("submission", method="form-data-post", **submission_attrs)
            model_children.insert(0, submission_node)
        return node("model",  *model_children)

    def xml_instance(self):
        result = Section.xml_instance(self)
        result.setAttribute(u"id", self.id_string)

        #add instance xmlns attribute to the instance node
        if self.instance_xmlns:
            result.setAttribute(u"xmlns", self.instance_xmlns)

        if self.version:
            result.setAttribute(u"version", self.version)
        return result

    def _add_to_nested_dict(self, dicty, path, value):
        if len(path) == 1:
            dicty[path[0]] = value
            return
        if path[0] not in dicty:
            dicty[path[0]] = {}
        self._add_to_nested_dict(dicty[path[0]], path[1:], value)

    def _setup_translations(self):
        """
        set up the self._translations dict which will be referenced in the setup media and itext functions
        """
        self._translations = defaultdict(dict)
        for element in self.iter_descendants():
            for d in element.get_translations(self.default_language):
                self._translations[d['lang']][d['path']] = {"long" : d['text']}

        #This code sets up translations for choices in filtered selects.
        for list_name, choice_list in self.choices.items():
            for idx, choice in zip(range(len(choice_list)), choice_list):
                for choicePropertyName, choicePropertyValue in choice.items():
                    itextId = '-'.join(['static_instance', list_name, str(idx)])
                    if isinstance(choicePropertyValue, dict):
                        for mediatypeorlanguage, value in choicePropertyValue.items():
                            if isinstance(value, dict):
                                for langauge, value in value.items():
                                    self._add_to_nested_dict(self._translations, [langauge, itextId, mediatypeorlanguage], value)
                            else:
                                if choicePropertyName == 'media':
                                    self._add_to_nested_dict(self._translations, [self.default_language, itextId, mediatypeorlanguage], value)
                                else:
                                    self._add_to_nested_dict(self._translations, [mediatypeorlanguage, itextId, 'long'], value)
                    elif choicePropertyName == 'label':
                        self._add_to_nested_dict(self._translations, [self.default_language, itextId, 'long'], choicePropertyValue)

    def _add_empty_translations(self):
        """
        Adds translations so that every itext element has the same elements accross every language.
        When translations are not provided "-" will be used.
        This disables any of the default_language fallback functionality.
        """
        paths = {}
        for lang, translation in self._translations.items():
            for path, content in translation.items():
                paths[path] = paths.get(path, set()).union(content.keys())

        for lang, translation in self._translations.items():
            for path, content_types in paths.items():
                if path not in self._translations[lang]:
                    self._translations[lang][path] = {}
                for content_type in content_types:
                    if content_type not in self._translations[lang][path]:
                        self._translations[lang][path][content_type] = u"-"

    def _setup_media(self):
        """
        Traverse the survey, find all the media, and put in into the _translations data structure which looks like this:
        {language : {element_xpath : {media_type : media}}}
        It matches the xform nesting order.
        """
        if not self._translations:
            self._translations = defaultdict(dict)

        for survey_element in self.iter_descendants():

            translation_key = survey_element.get_xpath() + ":label"
            media_dict = survey_element.get(u"media")

            for media_type, possibly_localized_media in media_dict.items():

                if media_type not in SurveyElement.SUPPORTED_MEDIA:
                    raise PyXFormError("Media type: " + media_type + " not supported")

                localized_media = dict()

                if type(possibly_localized_media) is dict:
                    #media is localized
                    localized_media = possibly_localized_media
                else:
                    #media is not localized so create a localized version using the default language
                    localized_media = { self.default_language : possibly_localized_media }

                for language, media in localized_media.items():

                    #Create the required dictionaries in _translations, then add media as a leaf value:

                    if language not in self._translations:
                        self._translations[language] = {}

                    translations_language = self._translations[language]

                    if translation_key not in translations_language:
                        translations_language[translation_key] = {}

                    #if type(translations_language[translation_key]) is not dict:
                    #    translations_language[translation_key] = {"long" : translations_language[translation_key]}

                    translations_trans_key = translations_language[translation_key]

                    if media_type not in translations_trans_key:
                            translations_trans_key[media_type] = {}

                    translations_trans_key[media_type] = media

    def itext(self):
        """
        This function creates the survey's itext nodes from _translations
        @see _setup_media _setup_translations
        itext nodes are localized images/audio/video/text
        @see http://code.google.com/p/opendatakit/wiki/XFormDesignGuidelines
        """
        result = []
        for lang, translation in self._translations.items():
            if lang == self.default_language:
                result.append(node("translation", lang=lang, default=u"true()"))
                #result.append(node("translation", lang=lang))
            else:
                result.append(node("translation", lang=lang))

            for label_name, content in translation.items():
                itext_nodes = []
                label_type = label_name.partition(":")[-1]

                if type(content) is not dict: raise Exception()

                for media_type, media_value in content.items():

                    #There is a odk/jr bug where hints can't have a value for the "form" attribute.
                    #This is my workaround.
                    if label_type == u"hint":
                        value, outputInserted = self.insert_output_values(media_value)
                        itext_nodes.append(node("value", value, toParseString=outputInserted))
                        continue

                    if media_type == "long":
                        value, outputInserted = self.insert_output_values(media_value)
                        #I'm ignoring long types for now because I don't know how they are supposed to work.
                        #itext_nodes.append(node("value", value, form=media_type, toParseString=outputInserted))
                        itext_nodes.append(node("value", value, toParseString=outputInserted))
                    elif media_type == "image":
                        itext_nodes.append(node("value", "jr://images/" + media_value, form=media_type))
                    else:
                        itext_nodes.append(node("value", "jr://" + media_type + "/" + media_value, form=media_type))


                result[-1].appendChild(node("text", *itext_nodes, id=label_name))

        return node("itext", *result)

    def date_stamp(self):
        return self._created.strftime("%Y_%m_%d")

    def _to_pretty_xml(self):
        """
        I want the to_xml method to by default validate the xml we are
        producing.
        """
        # Hacky way of pretty printing xml without adding extra white
        # space to text
        # TODO: check out pyxml
        # http://ronrothman.com/public/leftbraned/xml-dom-minidom-toprettyxml-and-silly-whitespace/
        xml_with_linebreaks = self.xml().toprettyxml(indent='  ')
        text_re = re.compile('>\n\s+([^<>\s].*?)\n\s+</', re.DOTALL)
        output_re = re.compile('\n.*(<output.*>)\n(  )*')
        prettyXml = text_re.sub('>\g<1></', xml_with_linebreaks)
        inlineOutput = output_re.sub('\g<1>', prettyXml)
        inlineOutput = re.compile('<label>\s*\n*\s*\n*\s*</label>').sub('<label></label>', inlineOutput)
        return '<?xml version="1.0"?>\n' + inlineOutput

    def __unicode__(self):
        return "<survey name='%s' element_count='%s'>" % (self.name, len(self.children))

    def _setup_xpath_dictionary(self):
        self._xpath = {}
        for element in self.iter_descendants():
            if isinstance(element, Question) or isinstance(element, Section):
                if element.name in self._xpath:
                    self._xpath[element.name] = None
                else:
                    self._xpath[element.name] = element.get_xpath()

    def _var_repl_function(self, matchobj):
        """
        Given a dictionary of xpaths, return a function we can use to
        replace ${varname} with the xpath to varname.
        """
        name = matchobj.group(1)
        intro = "There has been a problem trying to replace ${%s} with the XPath to the survey element named '%s'." % (name, name)
        if name not in self._xpath:
            raise PyXFormError(intro + " There is no survey element with this name.")
        if self._xpath[name] is None:
            raise PyXFormError(intro + " There are multiple survey elements with this name.")
        return " " + self._xpath[name] + " "


    def insert_xpaths(self, text):
        """
        Replace all instances of ${var} with the xpath to var.
        """
        #bracketed_tag = r"\$\{(" + XFORM_TAG_REGEXP + r")\}"
        bracketed_tag = r"\$\{(.*?)\}"
        return re.sub(bracketed_tag, self._var_repl_function, unicode(text))

    def _var_repl_output_function(self,matchobj):
        """
        A regex substitution function that will replace
        ${varname} with an output element that has the xpath to varname.
        """
#        if matchobj.group(1) not in self._xpath:
#            raise PyXFormError("There is no survey element with this name.",
#                            matchobj.group(1))
        return '<output value="' + self._var_repl_function(matchobj) + '" />'

    def insert_output_values(self, text):
        """
        Replace all the ${variables} in text with xpaths.
        Returns that and a boolean indicating if there were any ${variables} present.
        """
        #There was a bug where escaping is completely turned off in labels where
        #variable replacement is used.
        #For exampke, `${name} < 3` causes an error but `< 3` does not.
        #This is my hacky fix for it, which does string escaping prior to variable replacement:
        from xml.dom.minidom import Text
        text_node = Text()
        text_node.data = text
        xml_text = text_node.toxml()

        bracketed_tag = r"\$\{(.*?)\}"
        # need to make sure we have reason to replace
        # since at this point < is &lt,
        # the net effect &lt gets translated again to &amp;lt;
        if unicode(xml_text).find('{') != -1:
            result = re.sub(
                bracketed_tag, self._var_repl_output_function,
                unicode(xml_text))
            return result, not result == xml_text
        return text, False

    def print_xform_to_file(self, path="", validate=True, warnings=None):
        """
        Print the xForm to a file and optionally validate it as well by throwing exceptions
        and adding warnings to the warnings array.
        """
        if warnings is None:
            warnings = []
        if not path:
            path = self._print_name + ".xml"
        fp = codecs.open(path, mode="w", encoding="utf-8")
        fp.write(self._to_pretty_xml())
        fp.close()
        if validate:
            warnings.extend(check_xform(path))

    def to_xml(self, validate=True, pretty=True):
        with tempfile.NamedTemporaryFile() as tmp:
            # this will throw an exception if the xml is not valid
            self.print_xform_to_file(tmp.name)
        return self._to_pretty_xml()

    def instantiate(self):
        """
        Instantiate as in return a instance of SurveyInstance for collected data.
        """
        from instance import SurveyInstance
        return SurveyInstance(self)



from utils import node
from survey_element import SurveyElement
from question_type_dictionary import QUESTION_TYPE_DICT
from errors import PyXFormError


class Question(SurveyElement):

    def validate(self):
        SurveyElement.validate(self)

        # make sure that the type of this question exists in the
        # question type dictionary.
        if self.type not in QUESTION_TYPE_DICT:
            raise PyXFormError(
                "Unknown question type '%s'." % self.type
                )

    def xml_instance(self):
        if self.get(u"default"):
            #survey = self.get_root()
            #return node(self.name, survey.insert_xpaths(unicode(self.get(u"default"))))
            return node(self.name, unicode(self.get(u"default")))
        return node(self.name)

    def xml_control(self):
        return None


class InputQuestion(Question):
    """
    This control string is the same for: strings, integers, decimals,
    dates, geopoints, barcodes ...
    """
    def xml_control(self):
        control_dict = self.control
        label_and_hint = self.xml_label_and_hint()
        control_dict['ref'] = self.get_xpath()
        result = node(**control_dict)
        if label_and_hint:
            for element in self.xml_label_and_hint():
                result.appendChild(element)
        return result

class TriggerQuestion(Question):

    def xml_control(self):
        control_dict = self.control
        if u"appearance" in control_dict:
            return node(
                u"trigger", ref=self.get_xpath(),
                appearance=control_dict[u"appearance"],
                *self.xml_label_and_hint()
                )
        else:
            return node(u"trigger",
                ref=self.get_xpath(),
                *self.xml_label_and_hint()
                )


class UploadQuestion(Question):
    def _get_media_type(self):
        return self.control[u"mediatype"]

    def xml_control(self):
        control_dict = self.control
        if u"appearance" in control_dict:
            return node(
                u"upload",
                ref=self.get_xpath(),
                mediatype=self._get_media_type(),
                appearance=control_dict[u"appearance"],
                *self.xml_label_and_hint()
                )
        else:
            return node(
                u"upload",
                ref=self.get_xpath(),
                mediatype=self._get_media_type(),
                *self.xml_label_and_hint()
                )


class Option(SurveyElement):

    def xml_value(self):
        return node(u"value", self.name)

    def xml(self):
        item = node(u"item")
        xml_label = self.xml_label()
        item.appendChild(self.xml_label())
        item.appendChild(self.xml_value())
        return item

    def validate(self):
        pass

#class MultipleChoiceQuestion(Question):
#
#    def xml_control(self):
#        assert self.bind[u"type"] in [u"select", u"select1"]
#        control_dict = self.control.copy()
#        control_dict['ref'] = self.get_xpath()
#        nodeset = "instance('" + self['itemset'] + "')/root/item"
#        choice_filter = self.get('choice_filter')
#        if choice_filter:
#            survey = self.get_root()
#            choice_filter = survey.insert_xpaths(choice_filter)
#            nodeset += '[' + choice_filter + ']'
#        result = node(**control_dict)
#        for element in self.xml_label_and_hint():
#            result.appendChild(element)
#        itemset_label_ref = "jr:itext(itextId)"
#        itemset_children = [node('value', ref='name'), node('label', ref=itemset_label_ref)]
#        result.appendChild(node('itemset', *itemset_children, nodeset=nodeset))
#        return result
#        
#class SelectOneQuestion(MultipleChoiceQuestion):
#    pass
    
class MultipleChoiceQuestion(Question):

    def __init__(self, *args, **kwargs):
        kwargs_copy = kwargs.copy()
        #Notice that choices can be specified under choices or children. I'm going to try to stick to just choices.
        #Aliases in the json format will make it more difficult to use going forward.
        choices = kwargs_copy.pop(u"choices", []) + \
            kwargs_copy.pop(u"children", [])
        Question.__init__(self, *args, **kwargs_copy)
        for choice in choices:
            self.add_choice(**choice)

    def add_choice(self, **kwargs):
        option = Option(**kwargs)
        self.add_child(option)

    def validate(self):
        Question.validate(self)
        descendants = self.iter_descendants()
        descendants.next() # iter_descendants includes self; we need to pop it 
        for choice in descendants: 
            choice.validate()

    def xml_control(self):
        assert self.bind[u"type"] in [u"select", u"select1"]

        control_dict = self.control.copy()
        control_dict['ref'] = self.get_xpath()
        result = node(**control_dict)
        for element in self.xml_label_and_hint():
            result.appendChild(element)
        survey = self.get_root()
        # itemset are only supposed to be strings, check to prevent the rare dicts that show up
        if self['itemset'] and isinstance( self['itemset'] , basestring):
            choice_filter = self.get('choice_filter')
            nodeset = "instance('" + self['itemset'] + "')/root/item"
            choice_filter = survey.insert_xpaths(choice_filter)
            if choice_filter:
                nodeset += '[' + choice_filter + ']'
            itemset_label_ref = "jr:itext(itextId)"
            itemset_children = [node('value', ref='name'), node('label', ref=itemset_label_ref)]
            result.appendChild(node('itemset', *itemset_children, nodeset=nodeset))
        else:
            for n in [o.xml() for o in self.children]:
                result.appendChild(n)
        return result


class SelectOneQuestion(MultipleChoiceQuestion):
    def __init__(self, *args, **kwargs):
        super(SelectOneQuestion, self).__init__(*args, **kwargs)
        self._dict[self.TYPE] = u"select one"



"""
xls2xform converts properly formatted Excel documents into XForms for
use with ODK Collect.
"""
import sys
import xls2json
import builder
import json


# Converter.
def xls2xform_convert():
    # Warnings.
    warnings = []

    json_survey = xls2json.parse_file_to_json(argv[1], warnings=warnings)
    survey = builder.create_survey_element_from_dict(json_survey)
    # Setting validate to false will cause the form not to be processed by
    # ODK Validate.
    # This may be desirable since ODK Validate requires launching a subprocess
    # that runs some java code.
    survey.print_xform_to_file(argv[2], validate=True, warnings=warnings)

    return warnings


if __name__ == '__main__':
    argv = sys.argv
    if len(argv) < 3:
        print __doc__
        print 'Usage:'
        print argv[0] + ' path_to_XLSForm output_path'
        print '--json    Output results in json format.'
    else:

        # --json flag present. Capture everything and report in JSON format.
        if '--json' in argv:
            # Store everything in a list just in case the user wants to output
            # as a JSON encoded string.
            response = {'code': None, 'message': None, 'warnings': []}

            try:
                response['warnings'] = xls2xform_convert()

                response['code'] = 100
                response['message'] = "Ok!"

                if response['warnings']:
                    response['code'] = 101
                    response['message'] = 'Ok with warnings.'

            except Exception as e:
                # Catch the exception by default.
                response['code'] = 999
                response['message'] = str(e)

            print json.dumps(response)

        # --json not present. Do not capture anything.
        else:
            warnings = xls2xform_convert()

            # Regular output. Just print.
            for w in warnings:
                print w
            print 'Conversion complete!'



import os
import glob
import utils
#from xlrd import open_workbook

from xls2json import SurveyReader

def _section_name(path_or_file_name):
    directory, filename = os.path.split(path_or_file_name)
    section_name, extension = os.path.splitext(filename)
    return section_name

def load_file_to_dict(path):
    """
    Takes a file path and loads it into a nested json dict following the format in json_form_schema.json
    The file may be a xls file or json file. If it is xls it is converted using xls2json.
    """
    if path.endswith(".json"):
        name = _section_name(path)
        return (name, utils.get_pyobj_from_json(path))
    else:
        name = _section_name(path)
        excel_reader = SurveyReader(path)
        return (name, excel_reader.to_json_dict())

def collect_compatible_files_in_directory(directory):
    """
    create a giant dict out of all the spreadsheets and json forms in the given directory
    """
    #sections = {}
    available_files = glob.glob(os.path.join(directory, "*.xls")) + \
                        glob.glob(os.path.join(directory, "*.json"))
    return dict([load_file_to_dict(f) for f in available_files])



"""
pyxform is a Python library designed to make authoring XForms for ODK
Collect easy.
"""

__version__ = '0.9.20'

from survey import Survey
from section import Section
from question import MultipleChoiceQuestion, InputQuestion, Question
from instance import SurveyInstance
from builder import SurveyElementBuilder, create_survey_from_xls, \
    create_survey_element_from_dict, create_survey, create_survey_from_path
from question_type_dictionary import QUESTION_TYPE_DICT
from xls2json import SurveyReader as ExcelSurveyReader

#This is what gets imported when someone imports pyxform



from distutils.core import setup

setup(
    name='pyxform',
    version='0.9.20',
    author='modilabs',
    author_email='info@modilabs.org',
    packages=['pyxform', 'pyxform.odk_validate'],
    package_dir={'pyxform': 'pyxform'},
    package_data={
        'pyxform': [
            'odk_validate/ODK_Validate.jar',
        ],
    },
    url='http://pypi.python.org/pypi/pyxform/',
    description='A Python package to create XForms for ODK Collect.',
    long_description=open('README.rst', 'rt').read(),
    install_requires=[
        'xlrd==0.8.0',
        'lxml==2.3.4',
    ],
)
